use std::{
    collections::HashMap,
    io::{BufRead, BufReader, Write},
    net::{TcpListener, TcpStream},
};

#[derive(PartialEq, Eq, Debug)]
enum RequestMethods {
    NULL = -1, // This is only to initialise the struct
    GET,
    HEAD,
}

#[derive(Debug)]
struct RequestLine {
    method: RequestMethods,
    target: String,
    version: String,
}

impl RequestLine {
    pub fn new() -> Self {
        Self {
            method: RequestMethods::NULL,
            target: String::new(),
            version: String::new(),
        }
    }

    // https://datatracker.ietf.org/doc/html/rfc9110#name-methods
    pub fn is_valid_method(method: &String) -> bool {
        if method.trim().is_empty() {
            return false;
        }

        // Only GET and HEAD are required, the rest is optional
        if [
            "GET", "HEAD", /*, "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE",*/
        ]
        .contains(&method.trim())
        {
            return true;
        } else {
            return false;
        }
    }

    // TODO: make the checks less shit and actually correct
    // https://datatracker.ietf.org/doc/html/rfc9112#name-request-target
    pub fn is_valid_target(target: &String) -> bool {
        if target.trim().is_empty() {
            return false;
        }

        // origin-form
        if target.starts_with("/") {
            if target.contains("?") && target.split("?").count() != 2 {
                return false;
            }
            return true;
        }

        // absolute-form
        if target.starts_with("http://") {
            return true;
        }

        // authority-form
        if target.contains(":") && target.split(":").count() == 2 {
            return true;
        }

        // asterisk-form
        if target.trim() == "*" {
            return true;
        }

        return false;
    }

    pub fn is_valid_version(version: &String) -> bool {
        if version.trim().is_empty() {
            return false;
        }

        let http_name = version.trim();

        if http_name.starts_with("HTTP/") {
            let version_numbers = http_name.trim_start_matches("HTTP/");
            let version_numbers = version_numbers.split(".").collect::<Vec<&str>>();
            if version_numbers.len() != 2 {
                return false;
            }
            let major = match version_numbers[0].parse::<u8>() {
                Ok(val) => val,
                Err(_) => {
                    return false;
                }
            };
            let minor = match version_numbers[1].parse::<u8>() {
                Ok(val) => val,
                Err(_) => {
                    return false;
                }
            };

            if major <= 9 && minor <= 9 {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }
}

fn parse_start_line(input: String) -> Option<RequestLine> {
    if input.ends_with(" ") {
        return None;
    }

    let mut start_line = RequestLine::new();
    let vec = input.trim().split(" ").collect::<Vec<&str>>();

    if vec.len() != 3 {
        return None;
    }

    // start_line.method will remain RequestMethods::NULL if it is not supported.
    let method = String::from(vec[0]);
    if RequestLine::is_valid_method(&method) {
        // TODO: Change to a switch-case if I ever support more methods

        if method == String::from("GET") {
            start_line.method = RequestMethods::GET;
        }
        if method == String::from("HEAD") {
            start_line.method = RequestMethods::HEAD;
        }
    }

    let target = String::from(vec[1]);
    if RequestLine::is_valid_target(&target) {
        start_line.target = target;
    }

    let version = String::from(vec[2]);
    if RequestLine::is_valid_version(&version) {
        if version.trim() == "HTTP/1.1" || version.trim() == "HTTP/1.0" {
            start_line.version = version;
        }
    }

    return Some(start_line);
}

fn parse_field_lines(reader: &mut BufReader<&mut TcpStream>) -> Option<HashMap<String, String>> {
    let mut line: String;
    let mut field_lines: HashMap<String, String> = HashMap::new();
    let mut is_first_line = true;

    // Read field-lines till I hit an empty line
    loop {
        line = String::new();
        reader.read_line(&mut line).unwrap();

        if line.starts_with(" ") && is_first_line {
            return None;
        }

        if !line.ends_with("\r\n") {
            return None;
        }

        if line.trim().is_empty() {
            break;
        }

        let field_line = line.split_once(":").expect("Shits fucked: {}");

        // Check if client has send more than one Host line
        if field_lines.contains_key(&String::from("Host")) && field_line.0 == "Host" {
            return None;
        }

        field_lines.insert(field_line.0.to_owned(), field_line.1.trim().to_owned());

        is_first_line = false;
    }

    if !field_lines.contains_key(&String::from("Host")) {
        return None;
    }

    return Some(field_lines);
}

fn response_builder(
    method: RequestMethods,
    status_line: String,
    field_lines: HashMap<String, String>,
    body: Vec<u8>,
) -> Vec<u8> {
    let mut response: Vec<u8> = vec![];

    // TODO: replaces with eiter Option<T> or Result<T, T>
    if status_line.is_empty() {
        return response;
    }

    for byte in status_line.as_bytes().iter() {
        response.push(*byte);
    }
    response.push(b'\r');
    response.push(b'\n');

    if !field_lines.is_empty() {
        for field_line in field_lines.iter() {
            for byte in field_line.0.as_bytes().iter() {
                response.push(*byte);
            }

            response.push(b':');
            response.push(b' ');

            for byte in field_line.1.as_bytes().iter() {
                response.push(*byte);
            }

            response.push(b'\r');
            response.push(b'\n');
        }
    }

    // Mandatory empty line between header and body
    response.push(b'\r');
    response.push(b'\n');

    if body.is_empty() {
        return response;
    }

    if method != RequestMethods::HEAD && method != RequestMethods::NULL {
        for byte in body.iter() {
            response.push(*byte);
        }
        response.push(b'\r');
        response.push(b'\n');
    }

    return response;
}

fn handle_request(mut stream: TcpStream) {
    let mut line = String::new();
    let mut reader = BufReader::new(&mut stream);
    reader.read_line(&mut line).unwrap();

    let mut counter = 0;
    let max_preceding_empty_lines = 1000;

    // Request can have one or many empty lines preceding the start-line and I will ignore these
    // I will also for now only allow up to 1000 of these empty lines before I abort
    loop {
        if counter > max_preceding_empty_lines {
            stream
                .write_all(
                    b"HTTP/1.1 400 Bad Request\r\n\r\nReceived too many preceding empty lines",
                )
                .unwrap();
            return;
        }

        // All lines MUST end with a CRLF
        if !line.ends_with("\r\n") {
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nLines must end with CRLF")
                .unwrap();
            return;
        }
        line = line.trim().into();

        if !line.is_empty() {
            break;
        }
        reader.read_line(&mut line).unwrap();
        counter += 1;
    }

    let start_line = match parse_start_line(line) {
        Some(val) => val,
        None => {
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nInvalid Header")
                .unwrap();
            return;
        }
    };
    dbg!(&start_line);

    if start_line.method == RequestMethods::NULL {
        stream
            .write_all(b"HTTP/1.1 501 Not Implemented\r\n\r\nServer currently only supports GET")
            .unwrap();
        return;
    }

    let field_lines = match parse_field_lines(&mut reader) {
        Some(val) => val,
        None => {
            stream
                .write_all(b"HTTP/1.1 400 Bad Request\r\n\r\nInvalid Header")
                .unwrap();
            return;
        }
    };
    dbg!(&field_lines);

    // TODO: Read the body
    // let mut body: Vec<u8> = vec![];
    // reader.read_to_end(&mut body).unwrap();
    // dbg!(&body);

    // TODO: Act upon the request

    // TODO: Figure out why this doesn't work'
    let header: HashMap<String, String> = HashMap::new();
    let body: Vec<u8> = vec![];

    let response = response_builder(
        start_line.method,
        String::from("HTTP/1.1 200 OK"),
        header,
        body,
    );

    stream.write_all(&response).unwrap();

    // stream
    //     .write_all(b"HTTP/1.1 200 OK\r\n\r\nHello, World!\r\n")
    //     .unwrap();
}

fn main() {
    let listener = TcpListener::bind("127.0.0.1:8080").unwrap();

    for stream in listener.incoming() {
        let stream = stream.unwrap();
        handle_request(stream)
    }
}
